from curses import KEY_CANCEL
from dataclasses import dataclass
from random import randint, random
from pydbus import SessionBus
from pydbus.generic import signal
from gi.repository import GLib
import himitsu
import dhcrypto


loop = GLib.MainLoop()
bus = SessionBus()
hiq = himitsu.Client()


SESSIONS_NS = '/org/freedesktop/secrets/sessions/'


class Collection(object):
    '''
        <node>
            <interface name="org.freedesktop.Secret.Collection">
                <method name="Delete">
                    <arg type="o" name="prompt" direction="out" />
                </method>
                <method name="SearchItems">
                    <arg type="a{ss}" name="attributes" direction="in" />
                    <arg type="ao" name="results" direction="out" />
                </method>
                <method name="CreateItem">
                    <arg type="a{sv}" name="properties" direction="in" />
                    <arg type="(oayays)" name="secret" direction="in" />
                    <arg type="b" name="replace" direction="in" />
                    <arg type="o" name="item" direction="out" />
                    <arg type="o" name="prompt" direction="out" />
                </method>
                <signal name="ItemCreated">
                    <arg type="o" name="item" />
                </signal>
                <signal name="ItemDeleted">
                    <arg type="o" name="item" />
                </signal>
                <signal name="ItemChanged">
                    <arg type="o" name="item" />
                </signal>
                <property type="ao" name="Items" access="read" />
                <property type="s" name="Label" access="readwrite" />
                <property type="b" name="Locked" access="read" />
                <property type="t" name="Created" access="read" />
                <property type="t" name="Modified" access="read" />
            </interface>
        </node>
    '''

    @property
    def Items(self):
        return []

    Label = 'default'
    Private = 'yes?'
    Locked = False
    Created = 0
    Modified = 0

    ItemCreated = signal()
    ItemDeleted = signal()
    ItemChanged = signal()

class Service(object):
    '''
        <node>
            <interface name="org.freedesktop.Secret.Service">
                <method name="OpenSession">
                    <arg type="s" name="algorithm" direction="in" />
                    <arg type="v" name="input" direction="in" />
                    <arg type="v" name="output" direction="out" />
                    <arg type="o" name="result" direction="out" />
                </method>
                <method name="CreateCollection">
                    <arg type="a{sv}" name="properties" direction="in" />
                    <arg type="s" name="alias" direction="in" />
                    <arg type="o" name="collection" direction="out" />
                    <arg type="o" name="prompt" direction="out" />
                </method>
                <method name="SearchItems">
                    <arg type="a{ss}" name="attributes" direction="in" />
                    <arg type="ao" name="unlocked" direction="out" />
                    <arg type="ao" name="locked" direction="out" />
                </method>
                <method name="Unlock">
                    <arg type="ao" name="objects" direction="in" />
                    <arg type="ao" name="unlocked" direction="out" />
                    <arg type="o" name="prompt" direction="out" />
                </method>
                <method name="Lock">
                    <arg type="ao" name="objects" direction="in" />
                    <arg type="ao" name="locked" direction="out" />
                    <arg type="o" name="Prompt" direction="out" />
                </method>
                <method name="LockService" />
                <method name="ChangeLock">
                    <arg type="o" name="collection" direction="in" />
                    <arg type="o" name="prompt" direction="out" />
                </method>
                <method name="GetSecrets">
                    <arg type="ao" name="items" direction="in" />
                    <arg type="o" name="session" direction="in" />
                    <arg type="a{o(oayays)}" name="secrets" direction="out" />
                </method>
                <method name="ReadAlias">
                    <arg type="s" name="name" direction="in" />
                    <arg type="o" name="collection" direction="out" />
                </method>
                <method name="SetAlias">
                    <arg type="s" name="name" direction="in" />
                    <arg type="o" name="collection" direction="in" />
                </method>
                <signal name="CollectionCreated">
                    <arg type="o" name="collection" />
                </signal>
                <signal name="CollectionDeleted">
                    <arg type="o" name="collection" />
                </signal>
                <signal name="CollectionChanged">
                    <arg type="o" name="collection" />
                </signal>
                <property type="ao" name="Collections" access="read" />
            </interface>
            <node name="collection" />
        </node>
    '''
    sessions = {}
    keycache = []

    def OpenSession(self, algorithm, input):
        print('OpenSession', algorithm, input)
        session_id = len(self.sessions)
        session_path = f'{SESSIONS_NS}{str(session_id)}'
        session = dhcrypto.Session()
        self.sessions[session_id] = session

        if algorithm == "plain":
            return (GLib.Variant('s', 'plain'), session_path)
        if algorithm == "dh-ietf1024-sha256-aes128-cbc-pkcs7":
            pubkey = int.from_bytes(input, 'big')
            session.set_client_public_key(pubkey)
            return (GLib.Variant('ay', session.get_my_public_key()), session_path)

        # if algorithm != "plain":
        #     # i hate this
        #     exc = NotSupported("org.freedesktop.DBus.Error.NotSupported")
        #     type(exc).__name__ = 'org.freedesktop.DBus.Error.NotSupported'
        #     raise exc
        raise "you're not supposed to be here."
        return (None, None)

    def LockService(self):
        pass

    def SearchItems(self, attributes):
        print('SearchItems', attributes)
        results = hiq.query(**attributes)
        print('SearchItems#results', results)

        unlocked = []
        locked = []

        for res in results:
            objID = None
            if res in self.keycache:
                objID = self.keycache.index(res)
            else:
                objID = len(self.keycache)
                self.keycache.append(res)
            objPath = f'/org/freedesktop/secrets/items/{str(objID)}'
            if res.has_secret():
                locked.append(objPath)
            else:
                unlocked.append(objPath)

        print('SearchItems#locked', locked)
        print('SearchItems#unlocked', unlocked)
        return (locked, unlocked)

    def GetSecrets(self, items, session_path):
        print('GetSecrets', items, session_path)
        ret = {}
        for path in items:
            key = self.keycache[int(path.replace('/org/freedesktop/secrets/items/', ''))]
            session = self.sessions[int(session_path.replace(SESSIONS_NS, ''))]

            dec_key = hiq.query_decrypt(**key.to_query())
            # uhhhhhh himitsu can return multiple keys here, narrow it down to whichever has a secret
            dec_key = next(filter(lambda k: k.has_secret(), dec_key))
            secret = dec_key.get_secret().encode('utf-8')

            # handle encryption if necessary
            iv, enc = session.encrypt(secret)

            ret[path] = (session_path, iv, enc, 'text/plain')
        print('GetSecrets#ret', ret)
        return ret

    @property
    def Collections(self):
        return ['/org/freedesktop/secrets/collection/default']

    DefaultCollection = '/org/freedesktop/secrets/collection/default'

    CollectionCreated = signal()
    CollectionDeleted = signal()
    CollectionChanged = signal()

def entrypoint():
    print('Starting kakushi...')
    bus.publish(
        "org.freedesktop.secrets", Service(),
        ('collection/default', Collection())
    )
    loop.run()

if __name__ == '__main__':
    entrypoint()
